/* --------------------------------------------------------------
    Signed integers with unlimited range (version 1.2).

    http://www.imach.uran.ru/cbignum

    Copyright 1999-2006 by Raul N.Shakirov, IMach of RAS(UB).
    All Rights Reserved.

    Permission has been granted to copy, distribute and modify
    software in any context without fee, including a commercial
    application, provided that the aforesaid copyright statement
    is present here as well as exhaustive description of changes.

    THE SOFTWARE IS DISTRIBUTED "AS IS". NO WARRANTY OF ANY KIND
    IS EXPRESSED OR IMPLIED. YOU USE AT YOUR OWN RISK. THE AUTHOR
    WILL NOT BE LIABLE FOR DATA LOSS, DAMAGES, LOSS OF PROFITS OR
    ANY OTHER KIND OF LOSS WHILE USING OR MISUSING THIS SOFTWARE.
-------------------------------------------------------------- */
#include <assert.h>
#include <limits.h>
#include <stdlib.h>
#include "Cbignum.h"
#define BITS    (CHAR_BIT * sizeof (long))
static  int     log2_BITS = cLongExactLog2 (BITS);

#ifdef  _MSC_VER
#pragma auto_inline (off)
#endif

static  cBigNumber      a_;
static  cBigNumber      b_;
static  cBigNumber      num_;
cBigNumber cBigNumberLastDivMod;
cBigNumber cBigNumberLastRootRm;
exblockstack<long> cBigTemp::stack;

inline size_t size_max (size_t n1, size_t n2)
{
  return ((n1 > n2)? n1 : n2);
}
inline size_t size_min (size_t n1, size_t n2)
{
  return ((n1 < n2)? n1 : n2);
}

cBigNumber& cBigNumber::setbits     (const cBigNumber& b)
{
  size_t n = cBigNumberWords (EXPTRTYPE(b));
  set (n > 0? cULongBits ((unsigned long)
                (b.code()[--n] ^ (b.hiword()>>(BITS-1)))): 0);
  return addmul ((long)n, BITS);
}

cBigNumber& cBigNumber::setbits     (long b)
{
  return set (cLongBits (b));
}

cBigNumber& cBigNumber::setexbits   (const cBigNumber& b)
{
  size_t n = cBigNumberExWords (EXPTRTYPE(b));
  set (cLongBits (b.code()[n - (b.length() == 0)]));
  return addmul ((long)n, BITS);
}

cBigNumber& cBigNumber::setexbits   (long b)
{
  return set (cLongExBits (b));
}

cBigNumber& cBigNumber::fit()
{
  if (length()==0) checkexpand (1);
  checkindex (cBigNumberFit (EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::tab()
{
  checkexpand (exmuladd (BITS, length(), BITS * 2 + 1));
  cBigNumberTab (EXPTRTYPE(*this));
  return *this;
}

cBigNumber& cBigNumber::gc()
{
  checkshrink (length());
  return *this;
}

cBigNumber& cBigNumber::pack()
{
  reallocate (length() + (length()==0) + 1);
  return *this;
}

cBigNumber& cBigNumber::operator ++()
{
  checkexpand (length() + 1);
  checkindex (cBigNumberMInc (EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::operator --()
{
  checkexpand (length() + 1);
  checkindex (cBigNumberMDec (EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::mul2()
{
  checkexpand (length() + 1);
  checkindex (cBigNumberMMul2 (EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::div2()
{
  checkindex (cBigNumberMDiv2 (EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::pow2()
{
  checkexpand (length() * 2 + 2);
  cBigNumberMul (EXPTRTYPE(*this), EXPTRTYPE(*this), EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::sqrt()
{
  swap (cBigNumberLastRootRm);
  checkexpand (cBigNumberLastRootRm.fit().length() / 2 + 1);
  if (cBigNumberLastRootRm.hiword() >= 0)
    cBigNumberMRmSqrt (EXPTRTYPE(cBigNumberLastRootRm), EXPTRTYPE(*this));
  else
    *this = 0;
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setxor (const cBigNumber& a, const cBigNumber& b)
{
  checkindex ((a.length() >= b.length())?
    (checkexpand (a.length()),
     cBigNumberXor (EXPTRTYPE(a), EXPTRTYPE(b), EXPTRTYPE(*this))):
    (checkexpand (b.length()),
     cBigNumberXor (EXPTRTYPE(b), EXPTRTYPE(a), EXPTRTYPE(*this))));
  return *this;
}

cBigNumber& cBigNumber::setxor (const cBigNumber& a, long b)
{
  checkexpand (a.length());
  checkindex (cBigNumberXor (EXPTRTYPE(a), EXPTRTYPE(_cBigLong(b)),
                             EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setxor (long a, const cBigNumber& b)
{
  checkexpand (b.length());
  checkindex (cBigNumberXor (EXPTRTYPE(b), EXPTRTYPE(_cBigLong(a)),
                             EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setxor (long a, long b)
{
  return *this = a ^ b;
}

cBigNumber& cBigNumber::setand (const cBigNumber& a, const cBigNumber& b)
{
  checkindex ((a.length() >= b.length())?
    (checkexpand (a.length()),
     cBigNumberAnd (EXPTRTYPE(a), EXPTRTYPE(b), EXPTRTYPE(*this))):
    (checkexpand (b.length()),
     cBigNumberAnd (EXPTRTYPE(b), EXPTRTYPE(a), EXPTRTYPE(*this))));
  return *this;
}

cBigNumber& cBigNumber::setand (const cBigNumber& a, long b)
{
  checkexpand (a.length());
  checkindex (cBigNumberAnd (EXPTRTYPE(a), EXPTRTYPE(_cBigLong(b)),
                             EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setand (long a, const cBigNumber& b)
{
  checkexpand (b.length());
  checkindex (cBigNumberAnd (EXPTRTYPE(b), EXPTRTYPE(_cBigLong(a)),
                             EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setand (long a, long b)
{
  return *this = a & b;
}

cBigNumber& cBigNumber::setor  (const cBigNumber& a, const cBigNumber& b)
{
  checkindex ((a.length() >= b.length())?
    (checkexpand (a.length()),
     cBigNumberOr (EXPTRTYPE(a), EXPTRTYPE(b), EXPTRTYPE(*this))):
    (checkexpand (b.length()),
     cBigNumberOr (EXPTRTYPE(b), EXPTRTYPE(a), EXPTRTYPE(*this))));
  return *this;
}

cBigNumber& cBigNumber::setor  (const cBigNumber& a, long b)
{
  checkexpand (a.length());
  checkindex (cBigNumberOr (EXPTRTYPE(a), EXPTRTYPE(_cBigLong(b)),
                            EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setor  (long a, const cBigNumber& b)
{
  checkexpand (b.length());
  checkindex (cBigNumberOr (EXPTRTYPE(b), EXPTRTYPE(_cBigLong(a)),
                            EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setor  (long a, long b)
{
  return *this = a | b;
}

cBigNumber& cBigNumber::setadd (const cBigNumber& a, const cBigNumber& b)
{
  checkindex ((a.length() >= b.length())?
    (checkexpand (a.length() + 1),
     cBigNumberAdd (EXPTRTYPE(a), EXPTRTYPE(b), EXPTRTYPE(*this))):
    (checkexpand (b.length() + 1),
     cBigNumberAdd (EXPTRTYPE(b), EXPTRTYPE(a), EXPTRTYPE(*this))));
  return *this;
}

cBigNumber& cBigNumber::setadd (const cBigNumber& a, long b)
{
  checkexpand (a.length() + 1);
  checkindex (cBigNumberAdd (EXPTRTYPE(a), EXPTRTYPE(_cBigLong(b)),
                             EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setadd (long a, const cBigNumber& b)
{
  checkexpand (b.length() + 1);
  checkindex (cBigNumberAdd (EXPTRTYPE(b), EXPTRTYPE(_cBigLong(a)),
                             EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setadd (long a, long b)
{
  checkexpand (2);
  checkindex (cBigNumberAdd (EXPTRTYPE(_cBigLong(a)), EXPTRTYPE(_cBigLong(b)),
                             EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setsub (const cBigNumber& a, const cBigNumber& b)
{
  checkindex ((a.length() >= b.length())?
    (checkexpand (a.length() + 1),
     cBigNumberSub  (EXPTRTYPE(a), EXPTRTYPE(b), EXPTRTYPE(*this))):
    (checkexpand (b.length() + 1),
     cBigNumberSubS (EXPTRTYPE(b), EXPTRTYPE(a), EXPTRTYPE(*this))));
  return *this;
}

cBigNumber& cBigNumber::setsub (const cBigNumber& a, long b)
{
  checkexpand (a.length() + 1);
  checkindex (cBigNumberSub (EXPTRTYPE(a), EXPTRTYPE(_cBigLong(b)),
                             EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setsub (long a, const cBigNumber& b)
{
  checkexpand (b.length() + 1);
  checkindex (cBigNumberSubS(EXPTRTYPE(b), EXPTRTYPE(_cBigLong(a)),
                             EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setsub (long a, long b)
{
  checkexpand (2);
  checkindex (cBigNumberSub (EXPTRTYPE(_cBigLong(a)), EXPTRTYPE(_cBigLong(b)),
                             EXPTRTYPE(*this)));
  return *this;
}

cBigNumber& cBigNumber::setmul (const cBigNumber& a, const cBigNumber& b)
{
  size_t la = a.length(), lb = b.length();
  checkexpand (la + lb + 2);
  if (la >= lb)
        cBigNumberMul (EXPTRTYPE(a), EXPTRTYPE(b), EXPTRTYPE(*this));
  else  cBigNumberMul (EXPTRTYPE(b), EXPTRTYPE(a), EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setmul (const cBigNumber& a, long b)
{
  checkexpand (a.length() + 3);
  cBigNumberMul (EXPTRTYPE(a), EXPTRTYPE(_cBigLong(b)), EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setmul (long a, const cBigNumber& b)
{
  checkexpand (b.length() + 3);
  cBigNumberMul (EXPTRTYPE(b), EXPTRTYPE(_cBigLong(a)), EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setmul (long a, long b)
{
  checkexpand (4);
  cBigNumberMul (EXPTRTYPE(_cBigLong(a)), EXPTRTYPE(_cBigLong(b)),
                 EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setdiv (const cBigNumber& a, const cBigNumber& b)
{
  return setdivmod (cBigNumberLastDivMod = a, b);
}

cBigNumber& cBigNumber::setdiv (const cBigNumber& a, long b)
{
  return setdivmod (cBigNumberLastDivMod = a, b);
}

cBigNumber& cBigNumber::setdiv (long a, const cBigNumber& b)
{
  return setdivmod (cBigNumberLastDivMod = a, b);
}

cBigNumber& cBigNumber::setdiv (long a, long b)
{
  return setdivmod (cBigNumberLastDivMod = a, b);
}

cBigNumber& cBigNumber::setmod (const cBigNumber& a, const cBigNumber& b)
{
  return (*this = a) %= b;
}

cBigNumber& cBigNumber::setmod (const cBigNumber& a, long b)
{
  return (*this = a) %= b;
}

cBigNumber& cBigNumber::setmod (long a, const cBigNumber& b)
{
  return (*this = a) %= b;
}

cBigNumber& cBigNumber::setmod (long a, long b)
{
  return (*this = a) %= b;
}

cBigNumber& cBigNumber::operator <<= (const cBigNumber& b)
{
  if (b.hiword() < 0) return (*this >>= (-b));
  num_.setdivmod (b_ = b, BITS);
  size_t n = (size_t)b_.loword();
  fit();
  checkexpand ((num_.length() <= 1 &&
               (unsigned long)num_.loword() < UINT_MAX - 2 - length())?
                      (size_t)num_.loword() + length() + 2: UINT_MAX);
  for (; n > 0; --n) cBigNumberMMul2 (EXPTRTYPE(*this));
  cBigNumberCopyShl (EXPTRTYPE(*this), (size_t)num_.loword(),
                     EXPTRTYPE(*this));
  checkindex (length());
  EXDEBUG (num_ = 0);
  return *this;
}

cBigNumber& cBigNumber::operator <<= (long b)
{
  if (b < 0) return (*this >>= (-b));
  int n = (int)(b % BITS); b /= BITS;
  fit();
  checkexpand ((unsigned long)b < UINT_MAX - 2 - length()?
                      (size_t)b + length() + 2: UINT_MAX);
  for (; n > 0; --n) cBigNumberMMul2 (EXPTRTYPE(*this));
  cBigNumberCopyShl (EXPTRTYPE(*this), (size_t)b, EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setshl (const cBigNumber& a, const cBigNumber& b)
{
  *this = a; return *this <<= b;
}

cBigNumber& cBigNumber::setshl (const cBigNumber& a, long b)
{
  *this = a; return *this <<= b;
}

cBigNumber& cBigNumber::setshl (long a, const cBigNumber& b)
{
  *this = a; return *this <<= b;
}

cBigNumber& cBigNumber::setshl (long a, long b)
{
  *this = a; return *this <<= b;
}

cBigNumber& cBigNumber::operator >>= (const cBigNumber& b)
{
  if (b.hiword() < 0) return (*this <<= (-b));
  num_.setdivmod (b_ = b, BITS);
  size_t n = (size_t)b_.loword();
  fit();
  for (; n > 0; --n)  cBigNumberMDiv2 (EXPTRTYPE(*this));
  cBigNumberCopyShr (EXPTRTYPE(*this),
                     (num_.length() <= 1 &&
                     (unsigned long)num_.loword() <= UINT_MAX)?
                            (size_t)num_.loword(): UINT_MAX,
                     EXPTRTYPE(*this));
  checkindex (length());
  EXDEBUG (num_ = 0);
  return *this;
}

cBigNumber& cBigNumber::operator >>= (long b)
{
  if (b < 0) return (*this <<= (-b));
  size_t n = (size_t) b % BITS; b /= BITS;
  fit();
  for (; n > 0; --n) cBigNumberMDiv2 (EXPTRTYPE(*this));
  cBigNumberCopyShr (EXPTRTYPE(*this),
                     (unsigned long)b <= UINT_MAX?
                            (size_t)b: UINT_MAX,
                     EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setshr (const cBigNumber& a, const cBigNumber& b)
{
  *this = a; return *this >>= b;
}

cBigNumber& cBigNumber::setshr (const cBigNumber& a, long b)
{
  *this = a; return *this >>= b;
}

cBigNumber& cBigNumber::setshr (long a, const cBigNumber& b)
{
  *this = a; return *this >>= b;
}

cBigNumber& cBigNumber::setshr (long a, long b)
{
  *this = a; return *this >>= b;
}

cBigNumber& cBigNumber::setpow_()
{
  if (b_.hiword() < 0)
  {
    a_.setdiv (1,a_); b_.neg();
  }
  num_.setbits (a_); num_ *= b_;
  if (log2_BITS >= 0)
    num_ >>= log2_BITS;
  else
    num_.setdivmod (*this = num_, BITS);
  size_t n = ((num_.length() <= 1 &&
              (unsigned long)num_.loword() <= UINT_MAX - 3)?
                    ((size_t)num_.loword() + 3): UINT_MAX);
  checkexpand (n); a_.checkexpand (n);
  cBigNumberPow (EXPTRTYPE(a_), EXPTRTYPE(b_), EXPTRTYPE(*this));
  checkindex (length());
  EXDEBUG (a_ = 0);
  EXDEBUG (b_ = 0);
  EXDEBUG (num_ = 0);
  return *this;
}

cBigNumber& cBigNumber::setpow (const cBigNumber& a, const cBigNumber& b)
{
  a_ = a; b_ = b; return setpow_();
}

cBigNumber& cBigNumber::setpow (const cBigNumber& a, long b)
{
  a_ = a; b_ = b; return setpow_();
}

cBigNumber& cBigNumber::setpow (long a, const cBigNumber& b)
{
  a_ = a; b_ = b; return setpow_();
}

cBigNumber& cBigNumber::setpow (long a, long b)
{
  a_ = a; b_ = b; return setpow_();
}

cBigNumber& cBigNumber::setpowmod_ (const cBigNumber &mod)
{
  if (b_.hiword() < 0)
  {
    a_.setdiv (1,a_); b_.neg();
  }
  size_t n = mod.length() * 2 + 1;
  checkexpand (n); a_.checkexpand (size_max (a_.length() + 1, n));
  cBigNumberPowMod (EXPTRTYPE(a_), EXPTRTYPE(b_), EXPTRTYPE(mod),
                    EXPTRTYPE(*this));
  checkindex (length());
  EXDEBUG (a_ = 0);
  EXDEBUG (b_ = 0);
  return *this;
}

cBigNumber& cBigNumber::setpowmod (const cBigNumber& a, const cBigNumber& b,
                                   const cBigNumber& mod)
{
  a_ = a; b_ = b; return setpowmod_ (mod);
}

cBigNumber& cBigNumber::setpowmod (const cBigNumber& a, long b,
                                   const cBigNumber& mod)
{
  a_ = a; b_ = b; return setpowmod_ (mod);
}

cBigNumber& cBigNumber::setpowmod (long a, const cBigNumber& b,
                                   const cBigNumber& mod)
{
  a_ = a; b_ = b; return setpowmod_ (mod);
}

cBigNumber& cBigNumber::setpowmod (long a, long b,
                                   const cBigNumber& mod)
{
  a_ = a; b_ = b; return setpowmod_ (mod);
}

cBigNumber& cBigNumber::setpowmod_ (long mod)
{
  if (b_.hiword() < 0)
  {
    a_.setdiv (1,a_); b_.neg();
  }
  checkexpand (3); a_.checkexpand (size_max (a_.length() + 1, 3));
  cBigNumberPowMod (EXPTRTYPE(a_), EXPTRTYPE(b_), EXPTRTYPE(_cBigLong(mod)),
                    EXPTRTYPE(*this));
  checkindex (length());
  EXDEBUG (a_ = 0);
  EXDEBUG (b_ = 0);
  return *this;
}

cBigNumber& cBigNumber::setpowmod (const cBigNumber& a, const cBigNumber& b,
                                   long mod)
{
  a_ = a; b_ = b; return setpowmod_ (mod);
}

cBigNumber& cBigNumber::setpowmod (const cBigNumber& a, long b,
                                   long mod)
{
  a_ = a; b_ = b; return setpowmod_ (mod);
}

cBigNumber& cBigNumber::setpowmod (long a, const cBigNumber& b,
                                   long mod)
{
  a_ = a; b_ = b; return setpowmod_ (mod);
}

cBigNumber& cBigNumber::setpowmod (long a, long b,
                                   long mod)
{
  a_ = a; b_ = b; return setpowmod_ (mod);
}

cBigNumber& cBigNumber::setsqrt (const cBigNumber& b)
{
  cBigNumberLastRootRm = b;
  checkexpand (cBigNumberLastRootRm.fit().length() / 2 + 1);
  if (cBigNumberLastRootRm.hiword() >= 0)
    cBigNumberMRmSqrt (EXPTRTYPE(cBigNumberLastRootRm), EXPTRTYPE(*this));
  else
    *this = 0;
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setsqrt (long b)
{
  checkexpand (1);
  cBigNumberLastRootRm = b;
  if (b >= 0)
    cBigNumberMRmSqrt (EXPTRTYPE(cBigNumberLastRootRm), EXPTRTYPE(*this));
  else
    *this = 0;
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setrandom (unsigned long (*p)(), unsigned long lBits)
{
  checkexpand ((size_t)(lBits / BITS + 1));
  cBigNumberRandom (p, lBits, EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::add (const cBigNumber& b, size_t k)
{
  checkexpand (size_max (length(), b.length() + k) + 1);
  cBigNumberMAddShl (EXPTRTYPE(*this), EXPTRTYPE(b), k);
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::add (long b, size_t k)
{
  checkexpand (size_max (length(), k + 1) + 1);
  cBigNumberMAddShl (EXPTRTYPE(*this), EXPTRTYPE(_cBigLong(b)), k);
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::sub (const cBigNumber& b, size_t k)
{
  checkexpand (size_max (length(), b.length() + k) + 1);
  cBigNumberMSubShl (EXPTRTYPE(*this), EXPTRTYPE(b), k);
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::sub (long b, size_t k)
{
  checkexpand (size_max (length(), k + 1) + 1);
  cBigNumberMSubShl (EXPTRTYPE(*this), EXPTRTYPE(_cBigLong(b)), k);
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::addmul (const cBigNumber& a, const cBigNumber& b)
{
  size_t la = a.length(),lb = b.length();
  checkexpand (size_max (length(), la + lb + 1) + 1);
  if (la >= lb)
        cBigNumberMAddMul (EXPTRTYPE(a), EXPTRTYPE(b), EXPTRTYPE(*this));
  else  cBigNumberMAddMul (EXPTRTYPE(b), EXPTRTYPE(a), EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::addmul (const cBigNumber& a, long b)
{
  checkexpand (size_max (length(), a.length() + 2) + 1);
  cBigNumberMAddMul (EXPTRTYPE(a), EXPTRTYPE(_cBigLong(b)),
                     EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::addmul (long a, const cBigNumber& b)
{
  checkexpand (size_max (length(), b.length() + 2) + 1);
  cBigNumberMAddMul (EXPTRTYPE(b), EXPTRTYPE(_cBigLong(a)),
                     EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::addmul (long a, long b)
{
  checkexpand (size_max (length(), 3) + 1);
  cBigNumberMAddMul (EXPTRTYPE(_cBigLong(a)), EXPTRTYPE(_cBigLong(b)),
                     EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::submul (const cBigNumber& a, const cBigNumber& b)
{
  size_t la = a.length(),lb = b.length();
  checkexpand (size_max (length(), la + lb + 1) + 1);
  if (la >= lb)
        cBigNumberMSubMul (EXPTRTYPE(a), EXPTRTYPE(b), EXPTRTYPE(*this));
  else  cBigNumberMSubMul (EXPTRTYPE(b), EXPTRTYPE(a), EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::submul (const cBigNumber& a, long b)
{
  checkexpand (size_max (length(), a.length() + 2) + 1);
  cBigNumberMSubMul (EXPTRTYPE(a), EXPTRTYPE(_cBigLong(b)),
                     EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::submul (long a, const cBigNumber& b)
{
  checkexpand (size_max (length(), b.length() + 2) + 1);
  cBigNumberMSubMul (EXPTRTYPE(b), EXPTRTYPE(_cBigLong(a)),
                     EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::submul (long a, long b)
{
  checkexpand (size_max (length(), 3) + 1);
  cBigNumberMSubMul (EXPTRTYPE(_cBigLong(a)), EXPTRTYPE(_cBigLong(b)),
                     EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setdivmod (cBigNumber& a, const cBigNumber& b)
{
  checkexpand (a.fit().length() + 2);
  cBigNumberMModDiv (EXPTRTYPE(a), EXPTRTYPE(b), EXPTRTYPE(*this));
  a.checkindex (a.length());
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setdivmod (cBigNumber& a, long b)
{
  checkexpand (a.fit().length() + 2);
  cBigNumberMModDiv (EXPTRTYPE(a), EXPTRTYPE(_cBigLong(b)),
                     EXPTRTYPE(*this));
  a.checkindex (a.length());
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setsqrtrm (cBigNumber& a)
{
  checkexpand (a.fit().length() / 2 + 1);
  if (a.hiword() >= 0)
    cBigNumberMRmSqrt (EXPTRTYPE(a), EXPTRTYPE(*this));
  else
    *this = 0;
  a.checkindex (a.length());
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::addmultab (const cBigNumber& a, const cBigNumber& b,
                                   size_t k)
{
  checkexpand (size_max (length(), a.length() + b.length() + k) + 1);
  assert (b.hiword() >= 0);
  cBigNumberMAddMulShlTab (EXPTRTYPE(a), EXPTRTYPE(b), k, EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::addmultab (const cBigNumber& a, long b, size_t k)
{
  checkexpand (size_max (length(), a.length() + k + 1) + 1);
  assert (b >= 0);
  cBigNumberMAddMulShlTab (EXPTRTYPE(a), EXPTRTYPE(_cBigLong(b)), k,
                           EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::submultab (const cBigNumber& a, const cBigNumber& b,
                                   size_t k)
{
  checkexpand (size_max (length(), a.length() + b.length() + k) + 1);
  assert (b.hiword() >= 0);
  cBigNumberMSubMulShlTab (EXPTRTYPE(a), EXPTRTYPE(b), k, EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::submultab (const cBigNumber& a, long b, size_t k)
{
  checkexpand (size_max (length(), a.length() + k + 1) + 1);
  assert (b >= 0);
  cBigNumberMSubMulShlTab (EXPTRTYPE(a), EXPTRTYPE(_cBigLong(b)), k,
                           EXPTRTYPE(*this));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setdivtab (const cBigNumber& a, const cBigNumber& b)
{
  cBigNumberLastDivMod = a;
  checkexpand (cBigNumberLastDivMod.fit().length() + 2);
  cBigNumberMModDivTab (EXPTRTYPE(cBigNumberLastDivMod), EXPTRTYPE(b),
                        EXPTRTYPE(*this));
  cBigNumberLastDivMod.checkindex (cBigNumberLastDivMod.length());
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setdivtab (long a, const cBigNumber& b)
{
  cBigNumberLastDivMod = a;
  checkexpand (3);
  cBigNumberMModDivTab (EXPTRTYPE(cBigNumberLastDivMod), EXPTRTYPE(b),
                        EXPTRTYPE(*this));
  cBigNumberLastDivMod.checkindex (cBigNumberLastDivMod.length());
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setmodtab (const cBigNumber& a, const cBigNumber& b)
{
  *this = a; fit();
  cBigNumberMModTab (EXPTRTYPE(*this), EXPTRTYPE(b));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setmodtab (long a, const cBigNumber& b)
{
  *this = a;
  cBigNumberMModTab (EXPTRTYPE(*this), EXPTRTYPE(b));
  checkindex (length());
  return *this;
}

cBigNumber& cBigNumber::setdivmodtab (cBigNumber& a, const cBigNumber& b)
{
  checkexpand (a.fit().length() + 2);
  cBigNumberMModDivTab (EXPTRTYPE(a), EXPTRTYPE(b), EXPTRTYPE(*this));
  a.checkindex (a.length());
  checkindex (length());
  return *this;
}

int     b_SPRP  (
        const cBigNumber& n,
        const cBigNumber& b
)
{
  if (n <= 3) return (n < 2? 0: 2);
  if ((n.loword() & 1) == 0) return 0;

  cBigNumber t = n;
  cBigNumber nMinus1 = --t;
  cBigNumber a = nMinus1.exbits();
  t >>= a;

  cBigNumber test; test.setpowmod (b, t, n);

  if (test == 1 || test == nMinus1) return 2;

  if (a)
  {
    while (--a)
    {
      test.pow2(); test %= n; if (test == nMinus1) return 2;
    }
  }

  return 0;
}

const char* cBigNumber::pszConverting   = "Converting";
const char* cBigNumber::pszFormatting   = "Formatting";

const char  cBigNumber::numtochar [2] [MAX_RADIX + 1] =
                       {{'0','1','2','3','4','5','6','7','8','9',
                         'a','b','c','d','e','f','x'},
                        {'0','1','2','3','4','5','6','7','8','9',
                         'A','B','C','D','E','F','X'}};

      char  cBigNumber::chartonum [UCHAR_MAX + 1] = { 0 };

const long  cBigNumber::maxradix[][4] = {{ 0L,  1},
                                         { 1L,  1},
                                { 1073741824L, 30},
                                { 1162261467L, 19},
                                { 1073741824L, 15},
                                { 1220703125L, 13},
                                {  362797056L, 11},
                                { 1977326743L, 11},
                                { 1073741824L, 10},
                                {  387420489L,  9},
                                { 1000000000L,  9},
                                {  214358881L,  8},
                                {  429981696L,  8},
                                {  815730721L,  8},
                                { 1475789056L,  8},
                                {  170859375L,  7},
                                {  268435456L,  7} };

#define MAX_NP_IN   ((sizeof (size_t))==2? 2: 9)
#define MAX_NP_OUT  ((sizeof (size_t))==2? 2: 9)
#define MAX_NP      (MAX_NP_IN > MAX_NP_OUT? MAX_NP_IN: MAX_NP_OUT)

exarray<cBigNumber> cBigNumber::powradix [MAX_RADIX + 1];

size_t  cBigNumber::SqrTab (
                unsigned radix,
                size_t len,
                size_t max_np
)
{
  assert (radix >= 2);
  assert (radix <= MAX_RADIX);
  assert (max_np <= MAX_NP);
  static size_t npow [MAX_RADIX + 1];

  if (npow [radix] == 0)
  {
    int i;
    for (i = 0; i <= UCHAR_MAX; i++)
    {
      chartonum [i] = MAX_RADIX;
    }
    for (i = 0; i < MAX_RADIX; i++)
    {
      chartonum [numtochar [0] [i]] = (char)i;
      chartonum [numtochar [1] [i]] = (char)i;
    }

    powradix [radix] [0] = maxradix [radix] [0];
    powradix [radix] [0].tab();
    powradix [radix] [1] = maxradix [radix] [0];
    powradix [radix] [1].pow2();
    powradix [radix] [1].tab();
    npow [radix] = 2;
  }

  size_t np = 0;
  while (np < max_np && powradix [radix] [np].length() * 4 < len)
  {
    if (++np >= npow [radix])
    {
      powradix [radix] [np] = powradix [radix] [np - 1];
      powradix [radix] [np].pow2();
      powradix [radix] [np].tab();
      ++ npow [radix];
    }
  }
  assert (np <= max_np);
  return np;
}

static cBigNumber           mod_;
static exvector<cBigNumber> stk_(MAX_NP);
static exvector<int>       kstk_(MAX_NP);

cBigNumber& cBigNumber::setunsign (
                const EXPTR(char) ps,
                unsigned radix,
                unsigned width
            )
{
  assert (ps);
  assert (radix >= 2);
  assert (radix <= MAX_RADIX);

  *this = 0;
  if (width != 0)
  {
    long lradix =       maxradix [radix] [0];
    int  nradix = (int) maxradix [radix] [1];
    int nr = (width - 1) / nradix;
    int n = width - nr++ * nradix;

    size_t np = SqrTab (radix, nr, MAX_NP_IN);
    EXPTR(cBigNumber) powr = EXPTRTYPE(powradix [radix]);

    EXPTR(cBigNumber)  mul = EXPTRTYPE(stk_);
    { for (size_t i = 1; i < np; i++) mul [i] = 0; }

    cTTY cProgress (cBigNumberProgress);
    size_t numlength = nr * sizeof (long);
    if (numlength < cTTY_numscaleK) numlength = 0;
    if (numlength != 0 && np == MAX_NP_IN)
      cProgress.showScaled (cBigNumber::pszConverting, 0, (long)numlength);

    long lhigh = 0;
    if ((nr & 1) != 0) goto odd;
    do
    {
      {
        long lb = 0;
        for(;;)
        {
          lb += chartonum [*ps++];
          assert (lb >= 0);
          if (--n == 0) break;
          lb *= radix;
        }
        n = nradix;
        lhigh = lb;
      }
      --nr;

odd:  {
        long lb = 0;
        for(;;)
        {
          lb += chartonum [*ps++];
          assert (lb >= 0);
          if (--n == 0) break;
          lb *= radix;
        }
        n = nradix;
        num_ = lb;
      }
      --nr;

      num_.addmul (lhigh, lradix);
      assert (num_ >= 0);

      for (size_t i = 1;; i++)
      {
        if (i < np)
        {
          if ((nr & (2<<i)-1) != 0)
          {
            num_.swap (mul [i]);
            break;
          }
          num_.addmultab (powr [i], mul [i]);
          assert (num_ >= 0);
        }
        else
        {
          if (numlength != 0 && i == MAX_NP_IN)
            cProgress.showScaled (cBigNumber::pszConverting,
              (long)numlength - nr * sizeof (long), (long)numlength);
          num_.addmultab (powr [i], *this);
          assert (num_ >= 0);
          num_.swap (*this);
          break;
        }
      }
    }
    while (nr != 0);
  }

  EXDEBUG (num_ = 0);
  return *this;
}

cBigNumber& cBigNumber::set (
                const char* psz,
                unsigned radix
            )
{
  assert (psz);
  assert (radix != 1);
  assert (radix <= MAX_RADIX);
  int sign = 0;
  unsigned c = *psz++;

  switch (c)
  {
    case '-': sign = 1;
    case '+': c = *psz++;
  }

  if (radix == 0)
  {
    radix = 10;
    if (c == '0')
    {
      c = *psz++; radix = 8;
      if (c == 'X' || c == 'x') { c = *psz++; radix = 16; }
    }
  }

  SqrTab (radix);
  int width = 0;
  for (;;)
  {
    if ((size_t)chartonum [c] >= radix) break;
    c = psz [width++];
  }
  --psz;

  setunsign (EXPTRTO(char,psz,width), radix, width);

  if (sign) this->neg();

  return *this;
}

char*   cBigNumber::toatmp (
                unsigned radix,
                unsigned width,
                long     fill
        ) const
{
  assert (radix != 1);
  assert (radix <= MAX_RADIX);
  if (radix == 0) radix = 10;
  static exarray<char> str_;
  size_t i = 0;
  num_ = *this; num_.fit();
  if (fill & cBigNumber_unsign) num_.unsign();

  register const EXPTR(char) ps = EXPTRTO (char,
                    numtochar [(fill & cBigNumber_uppercase) != 0],
                    sizeof (numtochar [0]));

  {
    long sign = num_.comp0();

    if (sign < 0)
    {
      str_ [i++] = '-'; num_.neg();
    }
    else if (sign > 0 && (fill & cBigNumber_showpos))
    {
      str_ [i++] = '+';
    }

    if (radix % 8 == 0 && (fill & cBigNumber_showbase))
    {
      str_ [i++] = '0';
      if (radix == 16) str_ [i++] = ps [16];
    }
  }

  size_t j = i;
  {
    size_t np = SqrTab (radix, num_.length(), MAX_NP_OUT);
    EXPTR(cBigNumber) powr = EXPTRTYPE(powradix [radix]);

    EXPTR(cBigNumber)  div = EXPTRTYPE(stk_);
    EXPTR(int)        kdiv = EXPTRTYPE(kstk_);

    cTTY cProgress (cBigNumberProgress);
    size_t numlength = num_.length() * sizeof (long);
    if (numlength < cTTY_numscaleK) numlength = 0;

    size_t npmax = np;
    for (;;)
    {
      if (np == npmax)
      {
        if (numlength != 0 && np == MAX_NP_OUT)
          cProgress.showScaled (cBigNumber::pszFormatting,
            (long)(numlength - num_.length() * sizeof (long)),
            (long)numlength);

        num_.swap (mod_);
        num_.setdivmodtab (mod_, powr [np]);

        if (num_.comp0() == 0)
        {
          if (np == 0) break;
          num_.swap (mod_);
          npmax = --np;
          numlength = 0;
          assert (kdiv [np] == 0);
          continue;
        }
      }
      else
      {
        assert (np > 0);
        assert (kdiv [np - 1] == 0);
        if (kdiv [np] == 0)
        {
          ++np; continue;
        }
        mod_.swap (div [np]);
        kdiv [np] = 0;
      }

      for (; np != 0; np--)
      {
        assert (kdiv [np - 1] == 0);
        kdiv [np - 1] = 1;
        div  [np - 1].setdivmodtab (mod_, powr [np - 1]);
      }

      {
        register long lnum;
        lnum = mod_.loword();
        for (;;)
        {
          register long knum = maxradix [radix] [1] - 1;
          do
          {
            str_ [i++] = ps [ (int) (lnum % radix) ];
            lnum /= radix;
          }
          while (--knum != 0);

          assert (lnum >= 0);
          assert (lnum < (long) radix);
          str_ [i++] = ps [(int)(lnum)];

          if (kdiv [0] == 0) break;
          lnum = div [0].loword();
          kdiv [0] = 0;
        }
      }

      np = (npmax != 0);
    }

    {
      long lnum = mod_.loword();
      if (lnum != 0)
      do
      {
        str_ [i++] = ps [(int) (lnum % radix)];
      }
      while ((lnum /= radix) != 0);
    }

    if (i == j) str_ [i++] = ps [0];
  }

  {
    if ((char)fill == '\0') fill = ps [0];
    while (i < width) str_ [i++] = (char)fill;
    str_ [i--] = '\0';
  }

  {
    register EXPTR(char) ps = EXPTRTYPE(str_);
    while (i > j)
    {
      char c = ps [j]; ps [j] = ps [i]; j++; ps [i] = c; i--;
    }
  }

  EXDEBUG (num_ = 0);
  EXDEBUG (mod_ = 0);
  return ((char*)(const char *)str_);
}

cTTY& operator << (cTTY& os, const cBigNumber& b)
{
  cTTY_h ph = os;
  if (ph) (*ph) (&os, b.toatmp(), 0, 0, 0);
  return os;
}

void operator <<= (cTTY& os, const cBigNumber& b)
{
  cTTY_h ph = os;
  if (ph) (*ph) (&os, b.toatmp(), 0, 0, '\r');
}

cHexDump::cHexDump (unsigned long num)
{
  buffer [BITS / 4] = 0;
  for (size_t i = BITS / 4; i > 0; i--)
  {
    buffer [i - 1] = cBigNumber::numtochar [0] [(size_t)num & 0xF];
    num >>= 4;
  }
}
