/* --------------------------------------------------------------
    Signed integers with unlimited range (version 1.2).
    Service functions in machine-independent form.

    http://www.imach.uran.ru/cbignum

    Copyright 1999-2007 by Raul N.Shakirov, IMach of RAS(UB).
    All Rights Reserved.

    Permission has been granted to copy, distribute and modify
    software in any context without fee, including a commercial
    application, provided that the aforesaid copyright statement
    is present here as well as exhaustive description of changes.

    THE SOFTWARE IS DISTRIBUTED "AS IS". NO WARRANTY OF ANY KIND
    IS EXPRESSED OR IMPLIED. YOU USE AT YOUR OWN RISK. THE AUTHOR
    WILL NOT BE LIABLE FOR DATA LOSS, DAMAGES, LOSS OF PROFITS OR
    ANY OTHER KIND OF LOSS WHILE USING OR MISUSING THIS SOFTWARE.
-------------------------------------------------------------- */
#include <assert.h>
#include <limits.h>
#include <stdlib.h>
#include "Cbignum.h"
#define BITS            (CHAR_BIT * sizeof (long))

#ifdef  _MSC_VER
#pragma auto_inline (off)
#endif

#define SHIFT_TAB_MIN   (3)
#define SHIFT_TAB_OPT   (((sizeof(size_t))<=2? 120: 7680)/\
                         (sizeof (long) * sizeof (long)) - 3)
#define SHIFT_TAB_HIGH  (SHIFT_TAB_OPT + SHIFT_TAB_OPT / 2)
#define SHIFT_TAB_MAX   ((EXSIZE_T_MAX & (size_t)0xFFFFFFFFUL)/\
                         (256 * sizeof (long) * sizeof (long)))
#if 0
#define SIMPLE_MUL
#endif

inline size_t size_max (size_t n1, size_t n2)
{
  return ((n1 > n2)? n1 : n2);
}

inline size_t size_min (size_t n1, size_t n2)
{
  return ((n1 < n2)? n1 : n2);
}

#ifdef  __cplusplus
extern "C" {
#endif

cTTY    cBigNumberMessages;
cTTY    cBigNumberProgress;

void    cBigNumberDump (
                const   EXPTR(long) p1
        )
{
  size_t n1 = (size_t)(*p1);
  for (; n1; n1--) cBigNumberMessages <<" "<< (const char*) cHexDump (p1 [n1]);
  cBigNumberMessages << "\n";
}

void    cBigNumberInfo (
                const   EXPTR(long) p1
        )
{
  size_t n1 = (size_t)(*p1);
  cBigNumberMessages << cHexDump (p1 [n1]);
  if (n1 > 1)
    cBigNumberMessages << "... (" << (long)(n1 * sizeof (long)) << " bytes)\n";
}

void    cBigNumberERange  (
                const   EXPTR(long) p1
        )
{
  cBigNumberMessages << "cBigNumber value out of range: ";
  cBigNumberInfo (p1);
  abort();
}

size_t  cBigNumberWords (
                const   EXPTR(long) p1
        )
{
  size_t n1 = (size_t)(*p1);
  if (n1)
  {
    long sign = p1 [n1] >> (BITS-1);
    do
    {
      if (p1 [n1] != sign) break;
    }
    while (--n1 != 0);
  }
  return (n1);
}

int     cLongBits (long num)
{
  int n = BITS;
  num ^= (num >> (BITS-1));
  do --n; while ((num <<= 1) > 0);
  return (n);
}

int     cULongBits (unsigned long num)
{
  int n = BITS;
  while ((long)num > 0) { num <<= 1; --n; }
  return (n);
}

long    cBigNumberBits (
                const   EXPTR(long) p1
        )

{
  size_t n = cBigNumberWords(p1);
  size_t n1 = (size_t)(*p1);
  if (n)
  {
    if ((long)n > (long)(LONG_MAX / BITS))
       cBigNumberERange (p1);
    return ((long)(n - 1) * BITS + cULongBits ((unsigned long)
                                    (p1 [n] ^ (p1 [n1] >> (BITS-1)))));
  }
  return (0);
}

size_t  cBigNumberExWords (
                const   EXPTR(long) p1
)
{
  size_t n1 = (size_t)(*p1);
  for (size_t n = 0; n < n1; n++)
  {
    if (p1 [n + 1] != 0) return (n);
  }
  return (0);
}

int     cLongExBits (long num)
{
  int n = 0;
  if (num) while ((num & 1L) == 0) { ++n; num >>= 1; }
  return (n);
}

int     cLongExactLog2 (long num)
{
  int n = cLongExBits (num);
  if ((num >> n) != 1) n = -1;
  return n;
}

long    cBigNumberExBits (
                const   EXPTR(long) p1
        )

{
  size_t n = cBigNumberExWords(p1);
  size_t n1 = (size_t)(*p1);
  if (n1)
  {
    if ((long)n > (long)((LONG_MAX - (BITS - 1)) / BITS))
      cBigNumberERange (p1);
    return ((long)n * BITS + cLongExBits (p1 [n + 1]));
  }
  return (0);
}

size_t  cBigNumberIsFit (
                const   EXPTR(long) p1
        )
{
  size_t n1 = (size_t)(*p1);

  if (n1 > 1) return ((p1 [n1] != (p1 [n1-1] >> (BITS-1))));

  return (n1);
}

size_t  _cBigNumberFit  (
                        EXPTR(long) p1
        )
{
  register size_t n1 = (size_t)(*p1);
  p1 += n1;
  register long pass;
  register long lp;

  if (n1)
  {
    lp = *p1;
    do continue;
    while (--n1 != 0 &&
        (pass = lp, lp = p1 [-1], --p1, pass == (lp >> (BITS-1))));
  }
  else { p1 [1] = 0; }

  return (n1 + 1);
}

size_t  _cBigNumberFitTo (
                        EXPTR(long) p1,
                        size_t n
        )
{
  register size_t n1 = (size_t)(*p1);
  register long pass = p1 [n1] >> (BITS-1);

  do
  {
     p1 [++n1] = pass;
  }
  while (n1 < n);

  return (n1);
}

int     cBigNumberComp  (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2
        )
{
  size_t n1 = (size_t)(*p1);
  size_t n2 = (size_t)(*p2);
  register long lp1 = 0;
  register long lp2 = 0;

  if (n1) lp1 = p1 [n1--];
  if (n2) lp2 = p2 [n2--];

  if (n1 == n2)
  {
    if (lp1 != lp2) goto NotSameSignWord;
  }
  else if (n1 > n2)
  {
    lp2 >>= (BITS-1); ++n2;
    if (lp1 != lp2) goto NotSameSignWord;
    if (n1 > n2)
    {
      do
      {
        lp1 = p1 [n1]; --n1;
        if (lp1 != lp2) goto NotSame;
      }
      while (n1 > n2);
    }
  }
  else
  {
    lp1 >>= (BITS-1); ++n1;
    if (lp1 != lp2) goto NotSameSignWord;
    if (n1 < n2)
    {
      do
      {
        lp2 = p2 [n2]; --n2;
        if (lp1 != lp2) goto NotSame;
      }
      while (n1 < n2);
    }
  }

  if (n1)
  {
    do
    {
      lp1 = p1 [n1]; lp2 = p2 [n1]; --n1;
      if (lp1 != lp2) goto NotSame;
    }
    while (n1);
  }

  return (0);
NotSame:
  return (((unsigned long)lp1 >= (unsigned long)lp2) * 2 - 1);
NotSameSignWord:
  return ((lp1 >= lp2) * 2 - 1);
}

int     cBigNumberCompHigh (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2
        )
{
  size_t n1 = (size_t)(*p1);
  size_t n2 = (size_t)(*p2);
  assert (n1 >= n2);
  assert (n2 > 0);

  register long lp1 = p1 [n1--];
  register long lp2 = p2 [n2--];
  assert (!n1 || lp1 != (p1 [n1] >> (BITS-1)));
  assert (!n2 || lp2 != (p2 [n2] >> (BITS-1)));

  if (lp1 != lp2) return ((lp1 >= lp2) * 2 - 1);

  if (n2)
  do
  {
    lp1 = p1 [n1--]; lp2 = p2 [n2--];
    if (lp1 != lp2)
      return (((unsigned long)lp1 >= (unsigned long)lp2) * 2 - 1);
  }
  while (n2);

  if (n1)
  do
  {
    if (p1 [n1--] != 0) return 1;
  }
  while (n1);

  return (0);
}

inline  long    cBigNumberNot0  (
                const EXPTR(long) p1
                )
{
  assert (cBigNumberIsFit (p1));
  size_t n1 = (size_t)(*p1);
  return (p1 [n1] | (long)(n1 >> 1));
}

void    cBigNumberCopy (
                const   EXPTR(long) p1,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p++ = *p1++);
  if (n1)
  do
    { *p++ = *p1++; }
  while (--n1 != 0);
}

void    cBigNumberCompl (
                const   EXPTR(long) p1,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p++ = *p1++);
  if (n1)
  do
    { *p++ = ~*p1++; }
  while (--n1 != 0);
}

void    cBigNumberCopyShl (
                const   EXPTR(long) p1,
                        size_t      k1,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1);
  if (n1 <= 1 && p1 [n1] == 0) k1 = 0;
  p1 += n1; *p = (long)(n1 + k1); p += (n1 + k1);
  if (n1)
  do
    { *p-- = *p1--; }
  while (--n1 != 0);
  if (k1)
  do
    { *p-- = 0; }
  while (--k1 != 0);
}

void    cBigNumberCopyShr (
                const   EXPTR(long) p1,
                        size_t      k1,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1);
  if (n1 > k1)
  {
    p1 += (k1 + 1);
    *p++ = (long)(n1 -= k1);
    do
      { *p++ = *p1++; }
    while (--n1 != 0);
  }
  else if (n1)
  {
    *p++ = 1; *p = p1 [n1] >> (BITS-1);
  }
}

void    cBigNumberCutOut (
                const   EXPTR(long) p1,
                        size_t      k1,
                        size_t      n,
                        EXPTR(long) p
        )
{
  assert ((size_t)(*p1) >= k1 + n);
  *p++ = (long)(n + 1);
  if (n > 0)
  {
    p1 += (k1 + 1);
    do
      { *p++ = *p1++; }
    while (--n != 0);
  }
  *p = 0;
}

size_t  _cBigNumberNeg (
                const   EXPTR(long) p1,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1++);
  p++;
  size_t n = n1;
  if (n)
  do
  {
    long lp;
    if ((*p = lp = -*p1) != 0)
    {
      if (--n != 0)
      {
        do
          { ++p1; ++p; *p = lp = ~*p1; }
        while (--n != 0);
        return (n1 - (lp == (p [-1] >> (BITS-1))));
      }
      if (lp != LONG_MIN) return (n1);
      p [1] = 0; return (n1 + 1);
    }
  }
  while ((++p1, ++p, --n) != 0);

  return (n1);
}

size_t  _cBigNumberMInc (
                        EXPTR(long) p1
        )
{
  size_t n1 = (size_t)(*p1++);
  size_t n = n1;

  if (n)
  {
    do
    {
      long lp;
      if ((*p1 = lp = *p1 + 1) != 0)
      {
        if (--n != 0)
        {
          if (--n != 0) return (n1);
          return (n1 - (p1 [1] == (lp >> (BITS-1))));
        }
        if (lp != LONG_MIN) return (n1);
        p1 [1] = 0; return (n1 + 1);
      }
    }
    while ((++p1, --n) != 0);
  }
  else { *p1 = 1; n1 = 1; }

  return (n1);
}

size_t  _cBigNumberMDec (
                        EXPTR(long) p1
        )
{
  size_t n1 = (size_t)(*p1++);
  size_t n = n1;

  if (n)
  {
    do
    {
      long lp;
      if ((*p1 = lp = *p1 - 1) != ~0)
      {
        if (--n != 0)
        {
          if (--n != 0) return (n1);
          return (n1 - (p1 [1] == (lp >> (BITS-1))));
        }
        if (lp != LONG_MAX) return (n1);
        p1 [1] = ~0; return (n1 + 1);
      }
    }
    while ((++p1, --n) != 0);
  }
  else { *p1 = -1; n1 = 1; }

  return (n1);
}

size_t  _cBigNumberMMul2 (
                        EXPTR(long) p1
        )
{
  size_t n1 = (size_t)(*p1++);
  size_t n = n1;
  long num = 0;
  if (n)
  do
  {
    long pass = (unsigned long)num >> (BITS-1);
    num = *p1;
    *p1++ = (num << 1) | pass;
  }
  while (--n != 0);

  num >>= (BITS-2);
  n1 += (num != (*p1 = num >> 1));

  return (n1);
}

size_t  _cBigNumberMul2 (
                const   EXPTR(long) p1,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1++);
  p++;
  long num = 0;
  size_t n = n1;

  if (n)
  do
  {
    long pass = (unsigned long)num >> (BITS-1);
    num = *p1++;
    *p++ = (num << 1) | pass;
  }
  while (--n != 0);

  num >>= (BITS-2);
  n1 += (num != (*p = num >> 1));

  return (n1);
}

size_t  _cBigNumberMDiv2 (
                        EXPTR(long) p1
        )
{
  size_t n1 = (size_t)(*p1);
  size_t n = n1;
  p1 += n;
  if (n)
  {
    if (--n != 0)
    {
      long num = *p1;
      n1 -= (num == (*p1-- = (num >> 1)));
      do
      {
        long pass = num << (BITS-1);
        num = *p1;
        *p1-- = ((unsigned long)num >> 1) | pass;
      }
      while (--n != 0);
    }
    else
    {
      (*p1) >>= 1;
    }
  }
  return (n1);
}

size_t  _cBigNumberDiv2 (
                const   EXPTR(long) p1,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1);
  size_t n = n1;
  p1 += n;
  p  += n;
  if (n)
  {
    if (--n != 0)
    {
      long num = *p1--;
      n1 -= (num == (*p-- = (num >> 1)));
      do
      {
        long pass = num << (BITS-1);
        num = *p1--;
        *p-- = ((unsigned long)num >> 1) | pass;
      }
      while (--n != 0);
    }
    else
    {
      *p = (*p1) >> 1;
    }
  }
  return (n1);
}

void    cBigNumberTab  (
                        EXPTR(long) p
        )
{
  size_t n = cBigNumberFit (p);
  int ns = BITS - 1;
  p [n + 1] = p [n] >> (BITS-1);
  n += 2;
  EXPTR(long) pp = p;
  do
  {
    pp += n;
    cBigNumberMul2 (p, pp);
    p = pp;
  }
  while (--ns > 0);
}

size_t  _cBigNumberXor  (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1++);
  size_t n2 = (size_t)(*p2++);
  p++;
  assert (n1 >= n2);
  if (n1 == 0) return 0;

  {
    register long lp2 = 0;

    {
      size_t n;
      if ((n = n2) != 0)
      {
        do { *p++ = *p1++ ^ (lp2 = *p2++); }
        while (--n != 0);
      }
    }

    {
      lp2 >>= (BITS-1);
      size_t n;
      if ((n = n1 - n2) != 0)
      {
        do { *p++ = *p1++ ^ lp2; }
        while (--n != 0);
      }
    }
    --p;
  }

  {
    register long pass;
    register long lp;
    register size_t n = n1;
    if (n)
    {
      lp = *p;
      do continue;
      while (--n != 0 &&
        (pass = lp, lp = p [-1], --p, pass == (lp >> (BITS-1))));
    }
    else { p [1] = 0; }

    return (n + 1);
  }
}

size_t  _cBigNumberAnd  (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1++);
  size_t n2 = (size_t)(*p2++);
  p++;
  assert (n1 >= n2);
  if (n1 == 0) return 0;

  {
    register long lp2 = 0;

    {
      size_t n;
      if ((n = n2) != 0)
      {
        do { *p++ = *p1++ & (lp2 = *p2++); }
        while (--n != 0);
      }
    }

    {
      lp2 >>= (BITS-1);
      size_t n;
      if ((n = n1 - n2) != 0)
      {
        do { *p++ = *p1++ & lp2; }
        while (--n != 0);
      }
    }
    --p;
  }

  {
    register long pass;
    register long lp;
    register size_t n = n1;
    if (n)
    {
      lp = *p;
      do continue;
      while (--n != 0 &&
        (pass = lp, lp = p [-1], --p, pass == (lp >> (BITS-1))));
    }
    else { p [1] = 0; }

    return (n + 1);
  }
}

size_t  _cBigNumberOr  (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1++);
  size_t n2 = (size_t)(*p2++);
  p++;
  assert (n1 >= n2);
  if (n1 == 0) return 0;

  {
    register long lp2 = 0;

    {
      size_t n;
      if ((n = n2) != 0)
      {
        do { *p++ = *p1++ | (lp2 = *p2++); }
        while (--n != 0);
      }
    }

    {
      lp2 >>= (BITS-1);
      size_t n;
      if ((n = n1 - n2) != 0)
      {
        do { *p++ = *p1++ | lp2; }
        while (--n != 0);
      }
    }
    --p;
  }

  {
    register long pass;
    register long lp;
    register size_t n = n1;
    if (n)
    {
      lp = *p;
      do continue;
      while (--n != 0 &&
        (pass = lp, lp = p [-1], --p, pass == (lp >> (BITS-1))));
    }
    else { p [1] = 0; }

    return (n + 1);
  }
}

size_t  _cBigNumberAdd  (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1++);
  size_t n2 = (size_t)(*p2++);
  p++;
  assert (n1 >= n2);
  if (n1 == 0) { --p1; ++n1; }

  register unsigned long pass = 0;
  {
    register long lp1 = 0;
    register long lp2 = 0;

    {
      size_t n;
      if ((n = n2) != 0)
      {
        do
        {
          lp1 = *p1++; lp2 = *p2++;
          pass = (~(*p++ = (pass >> (BITS-1)) + lp1 + lp2) & (lp1 ^ lp2))
                                                           | (lp1 & lp2);
        }
        while (--n != 0);
      }
    }

    {
      size_t n;
      if ((n = n1 - n2) != 0)
      {
        register const long *pp = p1;
        if ((lp2 & LONG_MIN) == 0)
        {
          do
          {
            lp1 = *pp++;
            pass = ~(*p++ = (pass >> (BITS-1)) + lp1) & lp1;
          }
          while (--n != 0);
        }
        else
        {
          do
          {
            lp1 = *pp++;
            pass = ~(*p++ = (pass >> (BITS-1)) + lp1 - 1) | lp1;
          }
          while (--n != 0);
        }
      }
    }
    --p;

    pass = (pass >> (BITS-1)) + ((lp1) >> (BITS-1)) + ((lp2) >> (BITS-1));
  }

  {
    register long lp = *p;
    register size_t n = n1;
    if (n && (long)pass == (lp >> (BITS-1)))
    {
      do continue;
      while (--n != 0 &&
        (pass = lp, lp = p [-1], --p, (long)pass == (lp >> (BITS-1))));
    }
    else { p [1] = pass; }

    return (n + 1);
  }
}

size_t  _cBigNumberSub  (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1++);
  size_t n2 = (size_t)(*p2++);
  p++;
  assert (n1 >= n2);
  if (n1 == 0) { --p1; ++n1; }

  register unsigned long pass = ~0UL;
  {
    register long lp1 = 0;
    register long lp2 = ~0;

    {
      size_t n;
      if ((n = n2) != 0)
      {
        do
        {
          lp1 = *p1++; lp2 = ~(*p2++);
          pass = (~(*p++ = (pass >> (BITS-1)) + lp1 + lp2) & (lp1 ^ lp2))
                                                           | (lp1 & lp2);
        }
        while (--n != 0);
      }
    }

    {
      size_t n;
      if ((n = n1 - n2) != 0)
      {
        register const long *pp = p1;
        if ((lp2 & LONG_MIN) == 0)
        {
          do
          {
            lp1 = *pp++;
            pass = ~(*p++ = (pass >> (BITS-1)) + lp1) & lp1;
          }
          while (--n != 0);
        }
        else
        {
          do
          {
            lp1 = *pp++;
            pass = ~(*p++ = (pass >> (BITS-1)) + lp1 - 1) | lp1;
          }
          while (--n != 0);
        }
      }
    }
    --p;

    pass = (pass >> (BITS-1)) + ((lp1) >> (BITS-1)) + ((lp2) >> (BITS-1));
  }

  {
    register long lp = *p;
    register size_t n = n1;
    if (n && (long)pass == (lp >> (BITS-1)))
    {
      do continue;
      while (--n != 0 &&
        (pass = lp, lp = p [-1], --p, (long)pass == (lp >> (BITS-1))));
    }
    else { p [1] = pass; }

    return (n + 1);
  }
}

size_t  _cBigNumberSubS (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1++);
  size_t n2 = (size_t)(*p2++);
  p++;
  assert (n1 >= n2);
  if (n1 == 0) { --p1; ++n1; }

  register unsigned long pass = ~0UL;
  {
    register long lp1 = ~0;
    register long lp2 = 0;

    {
      size_t n;
      if ((n = n2) != 0)
      {
        do
        {
          lp1 = ~(*p1++); lp2 = *p2++;
          pass = (~(*p++ = (pass >> (BITS-1)) + lp1 + lp2) & (lp1 ^ lp2))
                                                           | (lp1 & lp2);
        }
        while (--n != 0);
      }
    }

    {
      size_t n;
      if ((n = n1 - n2) != 0)
      {
        register const long *pp = p1;
        if ((lp2 & LONG_MIN) == 0)
        {
          do
          {
            lp1 = ~(*pp++);
            pass = ~(*p++ = (pass >> (BITS-1)) + lp1) & lp1;
          }
          while (--n != 0);
        }
        else
        {
          do
          {
            lp1 = ~(*pp++);
            pass = ~(*p++ = (pass >> (BITS-1)) + lp1 - 1) | lp1;
          }
          while (--n != 0);
        }
      }
    }
    --p;

    pass = (pass >> (BITS-1)) + ((lp1) >> (BITS-1)) + ((lp2) >> (BITS-1));
  }

  {
    register long lp = *p;
    register size_t n = n1;
    if (n && (long)pass == (lp >> (BITS-1)))
    {
      do continue;
      while (--n != 0 &&
        (pass = lp, lp = p [-1], --p, (long)pass == (lp >> (BITS-1))));
    }
    else { p [1] = pass; }

    return (n + 1);
  }
}

static  size_t  _cBigNumberSkip = 0;

size_t  cBigNumberSkipLow0 (
                const   EXPTR(long) p1
)
{
  size_t n1 = (size_t)(*p1);
  size_t n = 1;

  if (n1 != 0)
  do
  {
    if (p1 [n] != 0) break;
    n++;
  }
  while (--n1 != 0);

  return (_cBigNumberSkip = n);
}

#ifndef _CBIGNUM_ASM

void    cBigNumberMAddM (
                        EXPTR(long) p1,
                const   EXPTR(long) p2,
                        size_t k2
        )
{
  const EXPTR(long) pp = p1 + (size_t)(*p1);
  size_t n2 = (size_t)(*p2);
  assert (_cBigNumberSkip > 0);
  assert ((size_t)(*p1) >= n2 + k2);
  assert (n2 >= _cBigNumberSkip);
  p1 += k2;

  register unsigned long pass = 0;
  {
    const EXPTR(long) pp = p1 + n2;
    p1 += _cBigNumberSkip;
    p2 += _cBigNumberSkip;
    register long lp1;
    register long lp2;

    for (;;)
    {
      lp1 = *p1;        lp2 = *p2;
      pass += lp1;      lp1 ^= lp2;
      pass += lp2;      lp1 = ~lp1;
      *p1 = pass;       pass |= lp1;
      lp1 &= lp2;       pass = ~pass;
      ++p2;             pass |= lp1;
      if (p1 >= pp)     break;
      ++p1;             pass >>= (BITS-1);
    }
    if ((long)(lp2 ^= pass) >= 0) return;
  }

  {
    register long lp1;
    if ((long)pass < 0)
    {
      do
      {
        if (p1 >= pp) break;
        lp1 = *++p1;
      }
      while ((~(*p1 = lp1 + 1) & lp1) < 0);
    }
    else
    {
      do
      {
        if (p1 >= pp) break;
        lp1 = *++p1;
      }
      while ((~(*p1 = lp1 - 1) | lp1) >= 0);
    }
  }
}

void    cBigNumberMSubM (
                        EXPTR(long) p1,
                const   EXPTR(long) p2,
                        size_t k2
        )
{
  const EXPTR(long) pp = p1 + (size_t)(*p1);
  size_t n2 = (size_t)(*p2);
  assert (_cBigNumberSkip > 0);
  assert ((size_t)(*p1) >= n2 + k2);
  assert (n2 >= _cBigNumberSkip);
  p1 += k2;

  register long pass = 0;
  {
    const EXPTR(long) pp = p1 + n2;
    p1 += _cBigNumberSkip;
    p2 += _cBigNumberSkip;
    register long lp1;
    register long lp2;

    for (;;)
    {
      lp1 = *p1;        lp2 = *p2;
      pass += lp1;      lp1 ^= lp2;
      pass -= lp2;      lp2 = ~lp2;
      *p1 = pass;       pass |= lp1;
      lp1 &= lp2;       pass = ~pass;
      ++p2;             pass |= lp1;
      if (p1 >= pp)     break;
      ++p1;             pass >>= (BITS-1);
                        pass = ~pass;
    }
    if ((long)(lp2 ^= pass) >= 0) return;
  }

  {
    register long lp1;
    if ((long)pass < 0)
    {
      do
      {
        if (p1 >= pp) break;
        lp1 = *++p1;
      }
      while ((~(*p1 = lp1 + 1) & lp1) < 0);
    }
    else
    {
      do
      {
        if (p1 >= pp) break;
        lp1 = *++p1;
      }
      while ((~(*p1 = lp1 - 1) | lp1) >= 0);
    }
  }
}

size_t  _cBigNumberMSubD (
                        EXPTR(long) p1,
                const   EXPTR(long) p2,
                        size_t k2
        )
{
  size_t n2 = (size_t)(*p2);
  assert (_cBigNumberSkip > 0);
  assert ((p1 [(size_t)(*p1)] ^ p2 [n2]) >= 0);
  assert ((size_t)(*p1) >= n2 + k2);
  assert (n2 >= _cBigNumberSkip);
  p1 += k2;

  register long pass = 0;
  {
    const EXPTR(long) pp = p1 + n2;
    p1 += _cBigNumberSkip;
    p2 += _cBigNumberSkip;
    register long lp1;
    register long lp2;

    for (;;)
    {
      lp1 = *p1;        lp2 = *p2;
      pass += lp1;      lp1 ^= lp2;
      pass -= lp2;      lp2 = ~lp2;
      *p1 = pass;       lp2 &= lp1;
      if (p1 >= pp)     break;
      lp2 = ~lp2;       pass |= lp1;
      ++p1;             pass &= lp2;
      ++p2;             pass >>= (BITS-1);
    }
  }

  {
    register size_t n1 = n2 + k2;

    for (;;)
    {
      register long lp1;
      register long lp2;
      if (--n1 == 0)    break;
      lp2 = p1 [-1];    --p1;
      lp1 = pass;       pass = lp2;
      lp2 >>= (BITS-1);
      if (lp1 != lp2)   break;
    }
    return (n1 + 1);
  }
}

size_t  _cBigNumberMMul2M (
                        EXPTR(long) p1
        )
{
  size_t n1 = (size_t)(*p1);
  assert (_cBigNumberSkip > 0);
  assert (n1 >= _cBigNumberSkip);

  const EXPTR(long) pp = p1 + n1;
  p1 += _cBigNumberSkip;
  long num = 0;
  do
  {
    long pass = (unsigned long)num >> (BITS-1);
    num = *p1;
    *p1++ = (num << 1) | pass;
  }
  while (p1 <= pp);

  num >>= (BITS-2);
  n1 += (num != (*p1 = num >> 1));

  return (n1);
}

size_t  _cBigNumberMDiv2D (
                        EXPTR(long) p1
        )
{
  size_t n1 = (size_t)(*p1);
  assert (_cBigNumberSkip > 0);
  assert (n1 >= _cBigNumberSkip);
  assert ((p1 [_cBigNumberSkip] & 1L) == 0);

  const EXPTR(long) pp = p1 + _cBigNumberSkip;
  p1 += n1;
  long num = *p1;
  n1 -= (n1 > 1) & (num == (*p1-- = num >> 1));
  while (p1 >= pp)
  {
    long pass = num << (BITS-1);
    num = *p1;
    *p1-- = ((unsigned long)num >> 1) | pass;
  }

  return (n1);
}

#endif

void    cBigNumberMShrM (
                        EXPTR(long) p1
        )
{
  size_t n1 = (size_t)(*p1);
  assert (n1 > 1);
  *p1 = (long)--n1;
  do
    { p1 [1] = p1 [2]; p1++; }
  while ((long)--n1 > 0);
}

void    cBigNumberMShlD (
                        EXPTR(long) p1
        )
{
  size_t n1 = (size_t)(*p1);
  assert (n1 > 0);
  *p1 = (long)(n1 + 1);
  do
    { p1 [n1 + 1] = p1 [n1]; }
  while ((long)--n1 > 0);
  p1 [1] = 0;
}

void    cBigNumberMAddShl (
                        EXPTR(long) p1,
                const   EXPTR(long) p2,
                        size_t      k2
        )
{
  size_t n2 = (size_t)(*p2);

  _cBigNumberSkip = 1;
  cBigNumberFitTo (p1, n2 + k2 + 1);
  if (n2 > 0) cBigNumberMAddM (p1, p2, k2);

  cBigNumberFit (p1);
  EXDEBUG (_cBigNumberSkip = 0);
}

void    cBigNumberMSubShl (
                        EXPTR(long) p1,
                const   EXPTR(long) p2,
                        size_t      k2
        )
{
  size_t n2 = (size_t)(*p2);

  _cBigNumberSkip = 1;
  cBigNumberFitTo (p1, n2 + k2 + 1);
  if (n2 > 0) cBigNumberMSubM (p1, p2, k2);

  cBigNumberFit (p1);
  EXDEBUG (_cBigNumberSkip = 0);
}

static  cBigNumber   cBigBuf1;
static  cBigNumber   cBigBuf2;
static  cBigNumber   cBigBuf3;
static  cBigNumber   cBigBuf4;

void    cBigNumberMAddMulShl (
                        EXPTR(long) p1,
                const   EXPTR(long) p2,
                        size_t      k,
                        EXPTR(long) p
        )
{
  assert (p1 != p);
  assert (p2 != p);
  assert (p2 != p1);
  size_t n1 = (size_t)(*p1);
  size_t n2 = (size_t)(*p2);

  if (n1 >= cBigNumberSkipLow0 (p1) && n2 != 0)
  {
    cBigNumberFitTo (p, n1 + n2 + k + 1);
    --n2;
    size_t k2 = 0;
    for (; k2 != n2; k2++)
    {
      long num = p2 [k2 + 1];
      if (num == 0) continue;
      long mask = 1;
      for (;; cBigNumberMMul2M (p1))
      {
        if (num & mask) cBigNumberMAddM (p, p1, k + k2);
        if ((mask <<= 1) == 0) break;
      }
      cBigNumberMMul2M (p1);
      cBigNumberMShrM (p1);
    }
    {
      long num = p2 [k2 + 1];
      long mask = (num != 0);
      if (num > 0)
      do
        mask <<= 1;
      while ((num <<= 1) > 0);
      for (;; cBigNumberMMul2M (p1))
      {
        if (num & mask) cBigNumberMAddM (p, p1, k + k2);
        if ((mask <<= 1) == 0) break;
      }
    }
  }

  cBigNumberFit (p);
  EXDEBUG (_cBigNumberSkip = 0);
}

void    cBigNumberMSubMulShl (
                        EXPTR(long) p1,
                const   EXPTR(long) p2,
                        size_t      k,
                        EXPTR(long) p
        )
{
  assert (p1 != p);
  assert (p2 != p);
  assert (p2 != p1);
  size_t n1 = (size_t)(*p1);
  size_t n2 = (size_t)(*p2);

  if (n1 >= cBigNumberSkipLow0 (p1) && n2 != 0)
  {
    cBigNumberFitTo (p, n1 + n2 + k + 1);
    --n2;
    size_t k2 = 0;
    for (; k2 != n2; k2++)
    {
      long num = p2 [k2 + 1];
      if (num == 0) continue;
      long mask = 1;
      for (;; cBigNumberMMul2M (p1))
      {
        if (num & mask) cBigNumberMSubM (p, p1, k + k2);
        if ((mask <<= 1) == 0) break;
      }
      cBigNumberMMul2M (p1);
      cBigNumberMShrM (p1);
    }
    {
      long num = p2 [k2 + 1];
      long mask = (num != 0);
      if (num > 0)
      do
        mask <<= 1;
      while ((num <<= 1) > 0);
      for (;; cBigNumberMMul2M (p1))
      {
        if (num & mask) cBigNumberMSubM (p, p1, k + k2);
        if ((mask <<= 1) == 0) break;
      }
    }
  }

  cBigNumberFit (p);
  EXDEBUG (_cBigNumberSkip = 0);
}

void    cBigNumberMAddMulShlTab (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2,
                        size_t      k,
                        EXPTR(long) p
        )
{
  assert (p1 != p);
  assert (p2 != p);
  size_t n1 = (size_t)(*p1);
  size_t n2 = (size_t)(*p2);

  if (n1 >= cBigNumberSkipLow0 (p1))
  {
    cBigNumberFitTo (p, n1 + n2 + k + 1);
    n1 += 2;
#ifdef  SIMPLE_MUL
    for (size_t k2 = 0; k2 != n2; k2++)
    {
      long num = p2 [k2 + 1];
      long mask = 1;
      const EXPTR(long) pp1 = p1;
      for (;; pp1 += n1)
      {
        if (num & mask) cBigNumberMAddM (p, pp1, k + k2);
        if ((mask <<= 1) == 0) break;
      }
    }
#else
    long num2 = p2 [n2], num1 = num2;
    num2 <<= 1;
    long mode = 0;
    if ((num1 & num2) < 0)
    {
      cBigNumberMAddM (p, p1, k + n2);
      num1 = ~num1; num2 = ~num2; mode = ~0;
    }
    p1 += n1 * BITS; p1 -= n1;
    for (size_t k2 = n2; k2-- != 0;)
    {
      const EXPTR(long) pp1 = p1;
      for (size_t kb = BITS;; pp1 -= n1)
      {
        long num = num1;
        num1 = num2;
        if (kb != 2) num2 <<= 1;
        else num2 = (k2? p2 [k2]: 0L) ^ mode;
        if ((num | (num1 & num2) ) < 0)
        {
           if (mode == 0) cBigNumberMAddM (p, pp1, k + k2);
           else           cBigNumberMSubM (p, pp1, k + k2);
           if (num >= 0) { num1 = ~num1; num2 = ~num2; mode = ~mode; }
        }
        if (--kb == 0) break;
      }
    }
#endif
  }

  cBigNumberFit (p);
  EXDEBUG (_cBigNumberSkip = 0);
}

void    cBigNumberMSubMulShlTab (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2,
                        size_t      k,
                        EXPTR(long) p
        )
{
  assert (p1 != p);
  assert (p2 != p);
  size_t n1 = (size_t)(*p1);
  size_t n2 = (size_t)(*p2);

  if (n1 >= cBigNumberSkipLow0 (p1))
  {
    cBigNumberFitTo (p, n1 + n2 + k + 1);
    n1 += 2;
#ifdef  SIMPLE_MUL
    for (size_t k2 = 0; k2 != n2; k2++)
    {
      long num = p2 [k2 + 1];
      long mask = 1;
      const EXPTR(long) pp1 = p1;
      for (;; pp1 += n1)
      {
        if (num & mask) cBigNumberMSubM (p, pp1, k + k2);
        if ((mask <<= 1) == 0) break;
      }
    }
#else
    long num2 = p2 [n2], num1 = num2;
    num2 <<= 1;
    long mode = 0;
    if ((num1 & num2) < 0)
    {
      cBigNumberMSubM (p, p1, k + n2);
      num1 = ~num1; num2 = ~num2; mode = ~0;
    }
    p1 += n1 * BITS; p1 -= n1;
    for (size_t k2 = n2; k2-- != 0;)
    {
      const EXPTR(long) pp1 = p1;
      for (size_t kb = BITS;; pp1 -= n1)
      {
        long num = num1;
        num1 = num2;
        if (kb != 2) num2 <<= 1;
        else num2 = (k2? p2 [k2]: 0L) ^ mode;
        if ((num | (num1 & num2) ) < 0)
        {
           if (mode == 0) cBigNumberMSubM (p, pp1, k + k2);
           else           cBigNumberMAddM (p, pp1, k + k2);
           if (num >= 0) { num1 = ~num1; num2 = ~num2; mode = ~mode; }
        }
        if (--kb == 0) break;
      }
    }
#endif
  }

  cBigNumberFit (p);
  EXDEBUG (_cBigNumberSkip = 0);
}

void    cBigNumberMAddMul (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1);
  size_t n2 = (size_t)(*p2);

  cBigBuf1.checkexpand ((n2 < SHIFT_TAB_MIN)? (n1 + 3):
                        (n1 < SHIFT_TAB_HIGH)? ((n1 + 3) * BITS + 1):
                  size_max (n1 + 3, (SHIFT_TAB_HIGH + 3) * BITS + 1));
  cBigBuf2.checkexpand (n2 + 1);
  EXPTR(long) pp1 = EXPTRTYPE(cBigBuf1);
  EXPTR(long) pp2 = EXPTRTYPE(cBigBuf2);

  if (p2 [n2] >= 0)
  {
    cBigNumberCopy (p1, pp1);
    cBigNumberCopy (p2, pp2);
  }
  else
  {
    cBigNumberNeg (p1, pp1);
    cBigNumberNeg (p2, pp2);
    assert (pp2 [(size_t)(*pp2)] >= 0);
    *pp2 = (long)n2;
  }

  size_t k = 0;
  if (n1 > SHIFT_TAB_HIGH)
  {
    cBigBuf3.checkexpand ((SHIFT_TAB_OPT + 3) * BITS + 1);
    EXPTR(long) pp3 = EXPTRTYPE(cBigBuf3);
    do
    {
      cBigNumberCutOut (pp1, k, SHIFT_TAB_OPT, pp3);
      if (n2 >= SHIFT_TAB_MIN)
      {
        cBigNumberTab (pp3);
        cBigNumberMAddMulShlTab (pp3, pp2, k, p);
      }
      else
      {
        cBigNumberMAddMulShl (pp3, pp2, k, p);
      }
      k += SHIFT_TAB_OPT;
    }
    while (n1 > k + SHIFT_TAB_HIGH);
    cBigNumberCopyShr (pp1, k, pp1);
  }

  if (n2 >= SHIFT_TAB_MIN)
  {
    cBigNumberTab (pp1);
    cBigNumberMAddMulShlTab (pp1, pp2, k, p);
  }
  else
  {
    cBigNumberMAddMulShl (pp1, pp2, k, p);
  }

  EXDEBUG (cBigBuf1 = 0);
  EXDEBUG (cBigBuf2 = 0);
  EXDEBUG (cBigBuf3 = 0);
  return;
}

void    cBigNumberMSubMul (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1);
  size_t n2 = (size_t)(*p2);

  cBigBuf1.checkexpand ((n2 < SHIFT_TAB_MIN)? (n1 + 3):
                        (n1 < SHIFT_TAB_HIGH)? ((n1 + 3) * BITS + 1):
                  size_max (n1 + 3, (SHIFT_TAB_HIGH + 3) * BITS + 1));
  cBigBuf2.checkexpand (n2 + 1);
  EXPTR(long) pp1 = EXPTRTYPE(cBigBuf1);
  EXPTR(long) pp2 = EXPTRTYPE(cBigBuf2);

  if (p2 [n2] >= 0)
  {
    cBigNumberCopy (p1, pp1);
    cBigNumberCopy (p2, pp2);
  }
  else
  {
    cBigNumberNeg (p1, pp1);
    cBigNumberNeg (p2, pp2);
    assert (pp2 [(size_t)(*pp2)] >= 0);
    *pp2 = (long)n2;
  }

  size_t k = 0;
  if (n1 > SHIFT_TAB_HIGH)
  {
    cBigBuf3.checkexpand ((SHIFT_TAB_OPT + 3) * BITS + 1);
    EXPTR(long) pp3 = EXPTRTYPE(cBigBuf3);
    do
    {
      cBigNumberCutOut (pp1, k, SHIFT_TAB_OPT, pp3);
      if (n2 >= SHIFT_TAB_MIN)
      {
        cBigNumberTab (pp3);
        cBigNumberMSubMulShlTab (pp3, pp2, k, p);
      }
      else
      {
        cBigNumberMSubMulShl (pp3, pp2, k, p);
      }
      k += SHIFT_TAB_OPT;
    }
    while (n1 > k + SHIFT_TAB_HIGH);
    cBigNumberCopyShr (pp1, k, pp1);
  }

  if (n2 >= SHIFT_TAB_MIN)
  {
    cBigNumberTab (pp1);
    cBigNumberMSubMulShlTab (pp1, pp2, k, p);
  }
  else
  {
    cBigNumberMSubMulShl (pp1, pp2, k, p);
  }

  EXDEBUG (cBigBuf1 = 0);
  EXDEBUG (cBigBuf2 = 0);
  EXDEBUG (cBigBuf3 = 0);
  return;
}

void    cBigNumberMul (
                const   EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  size_t n1 = (size_t)(*p1);
  size_t n2 = (size_t)(*p2);

  cBigBuf1.checkexpand ((n2 < SHIFT_TAB_MIN)? (n1 + 3):
                        (n1 < SHIFT_TAB_HIGH)? ((n1 + 3) * BITS + 1):
                  size_max (n1 + 3, (SHIFT_TAB_HIGH + 3) * BITS + 1));
  cBigBuf2.checkexpand (n2 + 1);
  EXPTR(long) pp1 = EXPTRTYPE(cBigBuf1);
  EXPTR(long) pp2 = EXPTRTYPE(cBigBuf2);

  if (p2 [n2] >= 0)
  {
    cBigNumberCopy (p1, pp1);
    cBigNumberCopy (p2, pp2);
  }
  else
  {
    cBigNumberNeg (p1, pp1);
    cBigNumberNeg (p2, pp2);
    assert (pp2 [(size_t)(*pp2)] >= 0);
    *pp2 = (long)n2;
  }

  p [0] = 0;

  size_t k = 0;
  if (n1 > SHIFT_TAB_HIGH)
  {
    cBigBuf3.checkexpand ((SHIFT_TAB_OPT + 3) * BITS + 1);
    EXPTR(long) pp3 = EXPTRTYPE(cBigBuf3);
    do
    {
      cBigNumberCutOut (pp1, k, SHIFT_TAB_OPT, pp3);
      if (n2 >= SHIFT_TAB_MIN)
      {
        cBigNumberTab (pp3);
        cBigNumberMAddMulShlTab (pp3, pp2, k, p);
      }
      else
      {
        cBigNumberMAddMulShl (pp3, pp2, k, p);
      }
      k += SHIFT_TAB_OPT;
    }
    while (n1 > k + SHIFT_TAB_HIGH);
    cBigNumberCopyShr (pp1, k, pp1);
  }

  if (n2 >= SHIFT_TAB_MIN)
  {
    cBigNumberTab (pp1);
    cBigNumberMAddMulShlTab (pp1, pp2, k, p);
  }
  else
  {
    cBigNumberMAddMulShl (pp1, pp2, k, p);
  }

  EXDEBUG (cBigBuf1 = 0);
  EXDEBUG (cBigBuf2 = 0);
  EXDEBUG (cBigBuf3 = 0);
  return;
}

size_t  cBigNumberMaskDiv0;
void    cBigNumberDiv0()
{
  cBigNumberMaskDiv0 /= cBigNumberMaskDiv0;
}

void    cBigNumberMModDivTab (
                        EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  assert (cBigNumberIsFit (p1));
  assert (p1 != p);
  assert (p2 != p);
  assert (p2 != p1);
  size_t n1 = (size_t)(*p1);
  size_t n2 = (size_t)(*p2);
  size_t n = ((n1 > n2)? n1 - n2: 0) + 1;

  *p = (long)(n + 1); p += n; p [1] = 0;

  if (n2 >= cBigNumberSkipLow0 (p2))
  {
    assert ((p1 [n1] ^ p2 [n2]) >= 0);
    long lt = (p1 [n1] < 0) * 2 - 1;
    const EXPTR(long) pp2 = p2;
    n2 += 2;
    do
    {
      --n;
      pp2 += n2 * BITS;
      unsigned long mask = (unsigned long)
                           (1L << (BITS-1));
      *p = 0;
      do
      {
        pp2 -= n2;
        assert (*pp2 >= (long)(n2 - 2) && *pp2 < (long) n2);
        long diff = (long)n - *p1 + *pp2;
        if (diff <= 0 && (diff < 0 ||
            cBigNumberCompHigh (p1, pp2) != lt))
        {
          cBigNumberMSubD (p1, pp2, n);
          (*p) += mask;
        }
      }
      while ((mask >>= 1) != 0);
      --p;
    }
    while (n);
    assert (cBigNumberComp (p1, pp2) == lt);
  }
  else
  {
    cBigNumberDiv0();
    do { *p-- = 0; } while (--n);
  }

  cBigNumberFit (p);
  assert (p [(size_t)(*p)] >= 0);
  EXDEBUG (_cBigNumberSkip = 0);
}

void    cBigNumberMModDiv (
                        EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  assert (p1 != p);
  size_t n1 = (size_t)(*p1);
  size_t n2 = (size_t)(*p2);

  long nt = (long) ((n2 - n1 + SHIFT_TAB_MIN - 1) & (n2 - SHIFT_TAB_MAX));
  cBigBuf2.checkexpand ((size_t)((nt < 0)? ((n2 + 3) * BITS + 1): (n2 + 3)));
  EXPTR(long) pp2 = EXPTRTYPE(cBigBuf2);

  long signdiv;
  if ((p1 [n1] ^ p2 [n2]) >= 0)
    { cBigNumberCopy (p2, pp2); signdiv = 1; }
  else
    { cBigNumberNeg (p2, pp2); signdiv = -1; }

  if (nt < 0)
  {
    cBigNumberTab (pp2);
    cBigNumberMModDivTab (p1, pp2, p);
    if (signdiv < 0) cBigNumberNeg (p, p);
    EXDEBUG (cBigBuf2 = 0);
    return;
  }

  assert (cBigNumberIsFit (p1));
  n2 = cBigNumberFit (pp2);
  size_t n = ((n1 > n2)? n1 - n2: 0) + 1;

  *p = (long)(n + 1); p += n; p [1] = 0;

  if (n2 >= cBigNumberSkipLow0 (p2))
  {
    assert ((p1 [n1] ^ pp2 [n2]) >= 0);
    long lt = (p1 [n1] < 0) * 2 - 1;
    do
    {
      --n;
      *p = 0;
      cBigNumberMShlD (pp2);
      unsigned long mask = (1UL << (BITS-1));
      do
      {
        long diff = (long)cBigNumberMDiv2D (pp2)
                  + (long)n - *p1;
        if (diff <= 0 && (diff < 0 ||
            cBigNumberCompHigh (p1, pp2) != lt))
        {
          cBigNumberMSubD (p1, pp2, n);
          (*p) += mask;
        }
      }
      while ((mask >>= 1) != 0);
      --p;
    }
    while (n);
    assert (cBigNumberComp (p1, pp2) == lt);
  }
  else
  {
    cBigNumberDiv0();
    do { *p-- = 0; } while (--n);
  }

  cBigNumberFit (p);
  assert (p [(size_t)(*p)] >= 0);
  if (signdiv < 0) cBigNumberNeg (p, p);

  EXDEBUG (_cBigNumberSkip = 0);
  EXDEBUG (cBigBuf2 = 0);
}

void    cBigNumberMModTab (
                        EXPTR(long) p1,
                const   EXPTR(long) p2
        )
{
  assert (cBigNumberIsFit (p1));
  assert (p1 != p2);
  size_t n1 = (size_t)(*p1);
  size_t n2 = (size_t)(*p2);
  size_t n = ((n1 > n2)? n1 - n2: 0) + 1;

  if (n2 >= cBigNumberSkipLow0 (p2))
  {
    assert ((p1 [n1] ^ p2 [n2]) >= 0);
    long lt = (p1 [n1] < 0) * 2 - 1;
    const EXPTR(long) pp2 = p2;
    n2 += 2;
    do
    {
      --n;
      pp2 += n2 * BITS;
      unsigned long mask = (1UL << (BITS-1));
      do
      {
        pp2 -= n2;
        assert (*pp2 >= (long)(n2 - 2) && *pp2 < (long) n2);
        long diff = (long)n - *p1 + *pp2;
        if (diff <= 0 && (diff < 0 ||
            cBigNumberCompHigh (p1, pp2) != lt))
        {
          cBigNumberMSubD (p1, pp2, n);
        }
      }
      while ((mask >>= 1) != 0);
    }
    while (n);
    assert (cBigNumberComp (p1, pp2) == lt);
  }
  else
  {
    cBigNumberDiv0();
  }

  EXDEBUG (_cBigNumberSkip = 0);
}

void    cBigNumberMMod (
                        EXPTR(long) p1,
                const   EXPTR(long) p2
        )
{
  size_t n1 = (size_t)(*p1);
  size_t n2 = (size_t)(*p2);

  long nt = (long) ((n2 - n1 + SHIFT_TAB_MIN - 1) & (n2 - SHIFT_TAB_MAX));
  cBigBuf2.checkexpand ((size_t)((nt < 0)? ((n2 + 3) * BITS + 1): (n2 + 3)));
  EXPTR(long) pp2 = EXPTRTYPE(cBigBuf2);

  if ((p1 [n1] ^ p2 [n2]) >= 0)
    cBigNumberCopy (p2, pp2);
  else
    cBigNumberNeg  (p2, pp2);

  if (nt < 0)
  {
    cBigNumberTab (pp2);
    cBigNumberMModTab (p1, pp2);
    EXDEBUG (cBigBuf2 = 0);
    return;
  }

  assert (cBigNumberIsFit (p1));
  n2 = cBigNumberFit (pp2);
  size_t n = ((n1 > n2)? n1 - n2: 0) + 1;

  if (n2 >= cBigNumberSkipLow0 (p2))
  {
    assert ((p1 [n1] ^ pp2 [n2]) >= 0);
    long lt = (p1 [n1] < 0) * 2 - 1;
    do
    {
      --n;
      cBigNumberMShlD (pp2);
      unsigned long mask = (1UL << (BITS-1));
      do
      {
        long diff = (long)cBigNumberMDiv2D (pp2)
                  + (long)n - *p1;
        if (diff <= 0 && (diff < 0 ||
            cBigNumberCompHigh (p1, pp2) != lt))
        {
          cBigNumberMSubD (p1, pp2, n);
        }
      }
      while ((mask >>= 1) != 0);
    }
    while (n);
    assert (cBigNumberComp (p1, pp2) == lt);
  }
  else
  {
    cBigNumberDiv0();
  }

  EXDEBUG (_cBigNumberSkip = 0);
  EXDEBUG (cBigBuf2 = 0);
}

void    cBigNumberPow (
                        EXPTR(long) p1,
                const   EXPTR(long) p2,
                        EXPTR(long) p
        )
{
  assert (p1 != p);
  assert (p2 != p);
  assert (p2 != p1);
  size_t n2 = (size_t)(*p2);
  assert (p2 [n2] >= 0);
  p2++;

  p [0] = 1; p [1] = 1;

  if (n2 != 0)
  {
    for (;;)
    {
      long num = *p2++;
      long mask = 1;
      if (--n2 == 0)
      {
        if (num == 0) break;
        do
          mask <<= 1;
        while ((num <<= 1) > 0);
      }
      for (;;)
      {
        if (num & mask)
          cBigNumberMul (p1, p, p);
        if ((mask <<= 1) == 0) break;
        cBigNumberMul (p1, p1, p1);
      }
      if (n2 == 0) break;
      cBigNumberMul (p1, p1, p1);
    }
  }
}

void    cBigNumberPowMod (
                        EXPTR(long) p1,
                const   EXPTR(long) p2,
                const   EXPTR(long) mod,
                        EXPTR(long) p
        )
{
  assert (p1 != p);
  assert (p2 != p);
  assert (p2 != p1);
  size_t n2 = (size_t)(*p2);
  assert (p2 [n2] >= 0);
  p2++;
  size_t nmod = (size_t)(*mod);

  cBigBuf4.checkexpand (exmuladd (BITS, nmod, BITS * 3 + 1));
  EXPTR(long) pmod = EXPTRTYPE(cBigBuf4);

  if (mod [nmod] >= 0)
    { cBigNumberCopy (mod, pmod); }
  else
    { cBigNumberNeg  (mod, pmod); }

  cBigNumberTab (pmod);
  if (cBigNumberNot0 (pmod) == 0) n2 = 0;

  p [0] = 1; p [1] = 1;
  cBigNumberMModTab (p, pmod);

  if (n2 != 0 && p [1] != 0)
  {

    long sign = p1 [(size_t)(*p1)];
    if (sign < 0)
    {
      cBigNumberNeg (p1, p1);
      assert (p1 [(size_t)(*p1)] >= 0);
      sign = (*p2 << (BITS-1));
    }
    cBigNumberMModTab (p1, pmod);

    for (;;)
    {
      long num = *p2++;
      long mask = 1;
      if (--n2 == 0)
      {
        if (num == 0) break;
        do
          mask <<= 1;
        while ((num <<= 1) > 0);
      }
      for (;;)
      {
        if (num & mask)
        {
          if (*p1 > *p) cBigNumberMul (p1, p, p);
          else          cBigNumberMul (p, p1, p);
          cBigNumberMModTab (p, pmod);
        }
        if ((mask <<= 1) == 0) break;
        cBigNumberMul (p1, p1, p1);
        cBigNumberMModTab (p1, pmod);
      }
      if (n2 == 0) break;
      cBigNumberMul (p1, p1, p1);
      cBigNumberMModTab (p1, pmod);
    }
    if (sign < 0) cBigNumberNeg (p, p);
  }

  EXDEBUG (cBigBuf4 = 0);
}

void    cBigNumberMRmSqrt (
                        EXPTR(long) p1,
                        EXPTR(long) p
        )
{
  assert (cBigNumberIsFit (p1));
  assert (p1 != p);
  size_t n1 = (size_t)(*p1);
  assert (p1 [n1] >= 0);

  cBigBuf2.checkexpand (n1 + 1);
  EXPTR(long) pp2 = EXPTRTYPE(cBigBuf2);
  *pp2 = 0;
  cBigNumberFitTo (pp2, n1);

  long nbit = (cBigNumberBits (p1) + 1) & ~1;
  if ((nbit -= 2) >= 0)
  {

    {
      size_t nw = (size_t)(nbit / BITS);
      long lw = (1L << (nbit % BITS));
      pp2 [nw + 1] += lw;
      assert (pp2 [(size_t)(*pp2)] >= 0);
      p1  [nw + 1] -= lw;
      assert (p1  [(size_t)(*p1)]  >= 0);
    }

    while ((nbit -= 2) >= 0)
    {
      size_t nw = (size_t)(nbit / BITS);
      long lw = (1L << (nbit % BITS));
      pp2 [nw + 1] += lw;
      _cBigNumberSkip = nw + 1;
      if (cBigNumberComp (p1, pp2) < 0)
      {
        pp2 [nw + 1] -= lw;
        cBigNumberMDiv2D (pp2);
      }
      else
      {
        cBigNumberMSubD (p1, pp2, 0);
        assert (p1 [(size_t)(*p1)] >= 0);
        assert (cBigNumberIsFit (p1));
        pp2 [nw + 1] -= lw;
        cBigNumberMDiv2D (pp2);
        pp2 [nw + 1] += lw;
      }
      assert (pp2 [(size_t)(*pp2)] >= 0);
      assert (cBigNumberIsFit (pp2));
    }
  }

  cBigNumberFit (pp2);
  cBigNumberCopy (pp2, p);

  assert (cBigNumberComp (
    (_cBigNumberSkip = 1, cBigNumberMMul2M (pp2), pp2), p1) >= 0);

  EXDEBUG (_cBigNumberSkip = 0);
}

void    cBigNumberRandom (
                unsigned long (*pfnRand)(),
                unsigned long lBits,
                        EXPTR(long) p
        )
{
  long   l1 = lBits % BITS;
  size_t n1 = (size_t)(lBits / BITS + 1);

  if (l1 != 0) l1 = (*pfnRand)() & ((unsigned long)~0L >> (BITS-(size_t)l1));
  p [0] = (long)n1;
  p [n1] = l1;
  while (--n1 != 0) p [n1] = (*pfnRand)();

  cBigNumberFit (p);
  assert (p [(size_t)(*p)] >= 0);
}

#ifdef  __cplusplus
}
#endif
